<'
# Copyright (c) 2012-2017 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Thu Aug  2 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <cerrno>
#include <cstdlib>
#include <cstdio>
#include <cstring>

#include <vector>

#include "ros/master.h"
#include "ros/topic_manager.h"

#include "<"$comp">/c/client.h"
#include "<"$comp">_msglib.h"
#include "private.h"


/* <"[--- ${comp}_init --------------------------------------------------]"> */

genom_client
genom_<"$comp">_client_init(int argc, char *argv[],
                   genom_event *ex, const void **exdetail)
{
  static char progname[] = "<"$comp">";
  std::vector<std::string> nodes;
  pthread_mutexattr_t attr;
  struct ros_options opts;
  char *largv[argc+1];
  genom_state_component state;
  genom_client h;
  int i, e;

  h = new genom_client_s;
  if (!h) {
    if (ex) *ex = genom_fatal;
    if (exdetail) *exdetail = NULL;
    return NULL;
  }
  h->raise = ros_client_raise;
  h->raised = ros_client_raised;
  h->data = &h->context_data;

  h->context_data.h = h;

  h->instance = NULL;
  h->node = NULL;
  h->connections = NULL;
  h->events = 0;

  h->ex = genom_ok;
  h->exdetail = NULL;
  h->exsize = 0;

  h->next = 0;
  pthread_mutexattr_init(&attr);
  pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
  pthread_mutex_init(&h->activelk, &attr);

  /* must shift args so that getopt() works (argv[0] must be progname) */
  largv[0] = progname;
  for(i=0; i<argc; i++) {
    largv[i+1] = argv[i];
  }

  if (ros_getopt(h, argc+1, largv, &opts)) {
    if (ex) *ex = h->ex;
    if (exdetail) *exdetail = h->exdetail;
    free(h);
    return NULL;
  }

  if (ros_init(h, argc+1, largv)) {
    if (ex) *ex = h->ex;
    if (exdetail) *exdetail = h->exdetail;
    pthread_mutex_destroy(&h->activelk);
    delete h;
    return NULL;
  }

  h->instance = strdup(opts.instance?opts.instance:"<"$comp">");
  h->topics_timeout = opts.topics_timeout;
  try {
    /* lookup server */
    if (!ros::master::getNodes(nodes)) {
      genom_mwerr_detail d;
      snprintf(d.what, sizeof(d.what), "cannot list running nodes");
      genom_mwerr(&d, h);
      goto error;
    }
    for(i=e=0; (unsigned)i<nodes.size(); i++) {
      if (!strcmp(h->instance, nodes[i].c_str()+1) && nodes[i][0] == '/') {
        e = 1;
        break;
      }
    }
    if (!e) {
      genom_mwerr_detail d;
      snprintf(d.what, sizeof(d.what),
               "%s node does not seem to be running", h->instance);
      genom_mwerr(&d, h);
      goto error;
    }

    h->node = new ros::NodeHandle;
  }
  catch (ros::Exception e) {
    genom_mwerr_detail d;
    snprintf(d.what, sizeof(d.what), "%s", e.what());
    genom_mwerr(&d, h);
    goto error;
  }
  h->connections = new struct connections;

  /* check digest */
  genom_<"$comp">_client_genom_state_init_data(&state);
  if (genom_<"$comp">_client_genom_state_port(h, &state)) goto error;

  if (strcmp(state.digest, "<"[$component digest]">")) {
    genom_incompatible_digest_detail d;
    strcpy(d.server.version, state.version);
    strcpy(d.server.date, state.date);
    strcpy(d.client.version, "<"[$component version]">");
    strcpy(d.client.date, "<"[clock format [clock sec]]">");
    genom_incompatible_digest(&d, h);
    genom_<"$comp">_client_genom_state_fini_data(&state);
    goto error;
  }
  genom_<"$comp">_client_genom_state_fini_data(&state);

  /* connect to services */
<'foreach s [$component services] {'>
<'  if {[catch {$s task}]} {'>
  h->connections-><"[$s name]">_ = h->node->serviceClient<
    genom::srv_<"[$s name]">::input, genom::srv_<"[$s name]">::output
    >("/" + std::string(h->instance) + "/<"[$s name]">", true);
<'  } else {'>
  h->connections-><"[$s name]">_ =
    new gnmix_<"[$s name]">(
      *h->node, "/" + std::string(h->instance) + "/<"[$s name]">");
<'  }'>
<'}'>

  /* check connections */
<'foreach s [$component services] {'>
<'  if {[catch {$s task}]} {'>
  if (!h->connections-><"[$s name]">_.exists()) {
    genom_mwerr_detail d;
    snprintf(d.what, sizeof(d.what),
             "cannot connect to service %s", "<"[$s name]">");
    genom_mwerr(&d, h);
    goto error;
  }
<'  } else {'>
  for(i = 0; i < 10 * h->topics_timeout; i++) {
    if (ros::TopicManager::instance()->getNumSubscribers(
          "/" + std::string(h->instance) + "/<"[$s name]">/goal") > 0 &&
        ros::TopicManager::instance()->getNumPublishers(
          "/" + std::string(h->instance) + "/<"[$s name]">/status") > 0) {
      i = 0;
      break;
    }
    usleep(100000/*100ms*/);
  }
  if (i) {
    genom_mwerr_detail d;
    snprintf(d.what, sizeof(d.what),
             "cannot connect to action %s", "<"[$s name]">");
    genom_mwerr(&d, h);
    goto error;
  }
<'  }'>
<'}'>

  return h;

error:
  if (ex) *ex = h->ex;
  if (exdetail) *exdetail = h->exdetail;
  if (h->connections) {
<'foreach s [$component services] {'>
<'  if {[catch {$s task}]} {'>
    h->connections-><"[$s name]">_.shutdown();
<'  } else {'>
    if (h->connections-><"[$s name]">_)
      delete h->connections-><"[$s name]">_;
<'  }'>
<'}'>
    delete h->connections;
  }
  if (h->instance) free(h->instance);
  if (h->node) delete h->node;
  pthread_mutex_destroy(&h->activelk);
  delete h;
  ros_fini();
  return NULL;
}


/* <"[--- ${comp}_fini --------------------------------------------------]"> */

void
genom_<"$comp">_client_fini(genom_client h)
{
  if (h) {
    if (h->connections) {
<'foreach s [$component services] {'>
<'  if {[catch {$s task}]} {'>
      h->connections-><"[$s name]">_.shutdown();
<'  } else {'>
      if (h->connections-><"[$s name]">_)
        delete h->connections-><"[$s name]">_;
<'  }'>
<'}'>
      delete h->connections;
    }
    if (h->instance) free(h->instance);
    if (h->node) delete h->node;
    pthread_mutex_destroy(&h->activelk);
    delete h;
    ros_fini();
  }
}


/* <"[--- ${comp}_context -----------------------------------------------]"> */

genom_context
genom_<"$comp">_client_context(genom_client h)
{
  return h;
}


/* <"[--- ${comp}_name --------------------------------------------------]"> */

const char *
genom_<"$comp">_client_instance(genom_client h)
{
  return h->instance;
}


/* <"[--- ${comp}_eventfd -----------------------------------------------]"> */

int
genom_<"$comp">_client_eventfd(genom_client)
{
  return ros_eventfd();
}


/* <"[--- ${comp}_service_info ------------------------------------------]"> */

const struct genom_service_info *
genom_<"$comp">_client_service_info(genom_client h, int rqstid)
{
  return ros_service_info(h, rqstid);
}


/* <"[--- ${comp}_done --------------------------------------------------]"> */

int
genom_<"$comp">_client_done(genom_client h, int rqstid)
{
  return ros_done(h, rqstid);
}


/* <"[--- ${comp}_wait --------------------------------------------------]"> */

genom_event
genom_<"$comp">_client_wait(genom_client h, int rqstid)
{
  return ros_waitterm(h, rqstid);
}


/* <"[--- ${comp}_clean -------------------------------------------------]"> */

genom_event
genom_<"$comp">_client_clean(genom_client h, int rqstid)
{
  return ros_clean(h, rqstid);
}


/* <"[--- ${comp}_doevents -----------------------------------------------]"> */

genom_event
genom_<"$comp">_client_doevents(genom_client h)
{
  return ros_doevents(h);
}
